"""
This module provides a class to read values from a modbus TCP PLC given the configuration.
"""

from typing import Callable
from enum import Enum
from math import ceil
from pymodbus.client import AsyncModbusTcpClient
from pymodbus.exceptions import ModbusException
from pymodbus.payload import BinaryPayloadDecoder as BPD
from pymodbus.pdu import ModbusResponse
from pymodbus.constants import Endian as ModbusEndian


class PLCReadException(Exception):
    """
    Raised when an error occurs while reading from a PLC.
    """

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)


class PLCReadConnectionError(PLCReadException):
    """
    Raised when an error occurs while connecting to a PLC.
    """


class PLCReadError(PLCReadException):
    """
    Raised when an error occurs while reading from a PLC.
    """


class PLCDecodeError(PLCReadException):
    """
    Raised when an error occurs while decoding data from a PLC.
    """


class PLCEndianness(Enum):
    """
    The endianness of the data in the PLC.
    """

    BIG = "big"
    LITTLE = "little"

    def as_modbus_endian(self) -> ModbusEndian:
        """
        Convert the endianness to a pymodbus endian.
        """
        return {
            PLCEndianness.BIG: ModbusEndian.BIG,
            PLCEndianness.LITTLE: ModbusEndian.LITTLE,
        }[self]


class PLCRegisterType(Enum):
    """
    The types of registers that can be read from a PLC.
    """

    COILS = "coils"
    DISCRETE_INPUTS = "discrete_inputs"
    INPUT_REGISTERS = "input_registers"
    HOLDING_REGISTERS = "holding_registers"


class PLCValueType(Enum):
    """
    The data types that can be read from a PLC.
    """

    UINT8 = "uint8"
    UINT16 = "uint16"
    UINT32 = "uint32"
    UINT64 = "uint64"
    INT8 = "int8"
    INT16 = "int16"
    INT32 = "int32"
    INT64 = "int64"
    FLOAT16 = "float16"
    FLOAT32 = "float32"
    FLOAT64 = "float64"
    CHAR = "char"
    STRING = "string"
    BOOL = "bool"

    def max_value(self):
        """
        The maximum value that can be represented by the data type.
        """
        return {
            PLCValueType.UINT8: 2**8 - 1,
            PLCValueType.UINT16: 2**16 - 1,
            PLCValueType.UINT32: 2**32 - 1,
            PLCValueType.UINT64: 2**64 - 1,
            PLCValueType.INT8: 2**7 - 1,
            PLCValueType.INT16: 2**15 - 1,
            PLCValueType.INT32: 2**31 - 1,
            PLCValueType.INT64: 2**63 - 1,
            PLCValueType.FLOAT16: 2**16 - 1,
            PLCValueType.FLOAT32: 2**32 - 1,
            PLCValueType.FLOAT64: 2**64 - 1,
            PLCValueType.CHAR: 2**8 - 1,
            PLCValueType.STRING: None,
            PLCValueType.BOOL: 1,
        }[self]

    def min_value(self):
        """
        The minimum value that can be represented by the data type.
        """
        return {
            PLCValueType.UINT8: 0,
            PLCValueType.UINT16: 0,
            PLCValueType.UINT32: 0,
            PLCValueType.UINT64: 0,
            PLCValueType.INT8: -(2**7),
            PLCValueType.INT16: -(2**15),
            PLCValueType.INT32: -(2**31),
            PLCValueType.INT64: -(2**63),
            PLCValueType.FLOAT16: -(2**16),
            PLCValueType.FLOAT32: -(2**32),
            PLCValueType.FLOAT64: -(2**64),
            PLCValueType.CHAR: 0,
            PLCValueType.STRING: None,
            PLCValueType.BOOL: 0,
        }[self]


class RegisterAddress:
    """
    A class to represent a register address.
    """

    def __init__(self, address: int):
        self._address = address

    @property
    def address(self) -> int:
        """
        The address of the register.
        """
        return self._address

    def __str__(self) -> str:
        return f"0x{format(self._address, '04x')}"

    def __repr__(self) -> str:
        return f"RegisterAddress(address={self._address})"


class PLCReader:
    """
    Reads values from a modbus TCP PLC given the configuration.

    Usage:

    ```py
    plc = PLCReader(host="127.0.0.1", port=502)
    await plc.connect()
    coil_value = await plc.read(
        address=0,
        register_type=PLCRegisterType.COILS,
    )
    discrete_input_value = await plc.read(
        address=0,
        register_type=PLCRegisterType.DISCRETE_INPUTS,
    )
    input_register_value = await plc.read(
        address=0,
        register_type=PLCRegisterType.INPUT_REGISTERS,
        data_type=PLCValueType.UINT16,
    )
    holding_register_value = await plc.read(
        address=0,
        register_type=PLCRegisterType.HOLDING_REGISTERS,
        data_type=PLCValueType.FLOAT32,
    )
    holding_register_string = await plc.read(
        address=0,
        register_type=PLCRegisterType.HOLDING_REGISTERS,
        data_type=PLCValueType.STRING,
        size=10,
    )
    plc.close()
    ```
    """

    REGISTER_VTYPE = {
        PLCRegisterType.COILS: PLCValueType.BOOL,
        PLCRegisterType.DISCRETE_INPUTS: PLCValueType.BOOL,
        PLCRegisterType.INPUT_REGISTERS: None,
        PLCRegisterType.HOLDING_REGISTERS: None,
    }
    VTYPE_SIZE_BYTES = {
        PLCValueType.UINT8: 1,
        PLCValueType.UINT16: 2,
        PLCValueType.UINT32: 4,
        PLCValueType.UINT64: 8,
        PLCValueType.INT8: 1,
        PLCValueType.INT16: 2,
        PLCValueType.INT32: 4,
        PLCValueType.INT64: 8,
        PLCValueType.FLOAT16: 2,
        PLCValueType.FLOAT32: 4,
        PLCValueType.FLOAT64: 8,
        PLCValueType.CHAR: 1,
        PLCValueType.STRING: None,
        PLCValueType.BOOL: 1,
    }

    REGISTER_DECODERS: dict[
        PLCRegisterType, Callable[[ModbusResponse, PLCEndianness, PLCEndianness], BPD]
    ] = {
        PLCRegisterType.COILS: lambda r, e, w: BPD.fromCoils(
            coils=r.bits, byteorder=PLCEndianness.BIG.value, _wordorder=w
        ),
        PLCRegisterType.DISCRETE_INPUTS: lambda r, e, w: BPD.fromCoils(
            coils=r.bits, byteorder=PLCEndianness.BIG.value, _wordorder=w
        ),
        PLCRegisterType.INPUT_REGISTERS: lambda r, e, w: BPD.fromRegisters(
            registers=r.registers, byteorder=e, wordorder=w
        ),
        PLCRegisterType.HOLDING_REGISTERS: lambda r, e, w: BPD.fromRegisters(
            registers=r.registers, byteorder=e, wordorder=w
        ),
    }

    def __init__(
        self,
        host: str = "127.0.0.1",
        port: int = 502,
        endianness: PLCEndianness = PLCEndianness.BIG,
        word_order: PLCEndianness = PLCEndianness.BIG,
    ):
        self._host = host
        self._port = port
        self._word_order = word_order
        self._endianness = endianness
        self.__client = AsyncModbusTcpClient(host=host, port=port)
        self.__readers = {
            PLCRegisterType.COILS: self.__client.read_coils,
            PLCRegisterType.DISCRETE_INPUTS: self.__client.read_discrete_inputs,
            PLCRegisterType.INPUT_REGISTERS: self.__client.read_input_registers,
            PLCRegisterType.HOLDING_REGISTERS: self.__client.read_holding_registers,
        }

    async def read(
        self,
        address: int,
        register_type: PLCRegisterType,
        data_type: PLCValueType = PLCValueType.UINT16,
        size: int = 1,
    ):
        """
        Read a value from the PLC.

        :param address: The address of the register to read.
        :param register_type: The type of register to read.
        :param data_type: The data type of the register. (Ignored for coils and discrete inputs)
        :param size: The size of a string in bytes. Ignored for all data types except string.
        :returns: The value of the register as the type given by `data_type`.
        :raises PLCReadError: If the register could not be read.
        :raises PLCDecodeError: If the register could not be decoded.
        """
        return await self.__read(
            address=address,
            read_fn=self.__readers[register_type],
            register_type=register_type,
            data_type=PLCReader.REGISTER_VTYPE[register_type] or data_type,
            size=PLCReader.VTYPE_SIZE_BYTES[data_type] or size,
        )

    async def __read(
        self,
        address: int,
        read_fn: callable,
        register_type: PLCRegisterType,
        data_type: PLCValueType,
        size: int,
    ) -> int | str | bool:
        try:
            response = await read_fn(
                address=address,
                count=ceil(size / 2),
            )
        except ModbusException as exc:
            raise PLCReadError("Connection failed during read.") from exc

        if response.isError():
            raise PLCReadError(
                f"Unable to process register {RegisterAddress(address)} using {read_fn.__name__}"
            )

        decoder = PLCReader.REGISTER_DECODERS[register_type](
            r=response,
            e=self._endianness.as_modbus_endian(),
            w=self._word_order.as_modbus_endian(),
        )
        decode_func = {
            PLCValueType.UINT8: decoder.decode_8bit_uint,
            PLCValueType.UINT16: decoder.decode_16bit_uint,
            PLCValueType.UINT32: decoder.decode_32bit_uint,
            PLCValueType.UINT64: decoder.decode_64bit_uint,
            PLCValueType.INT8: decoder.decode_8bit_int,
            PLCValueType.INT16: decoder.decode_16bit_int,
            PLCValueType.INT32: decoder.decode_32bit_int,
            PLCValueType.INT64: decoder.decode_64bit_int,
            PLCValueType.FLOAT16: decoder.decode_16bit_float,
            PLCValueType.FLOAT32: decoder.decode_32bit_float,
            PLCValueType.FLOAT64: decoder.decode_64bit_float,
            PLCValueType.CHAR: lambda: decoder.decode_string(size=1),
            PLCValueType.STRING: lambda: decoder.decode_string(size=size),
            PLCValueType.BOOL: lambda: response.bits[0]
            if self._endianness == PLCEndianness.BIG
            else response.bits[-1],
        }[data_type]
        try:
            decode_value = decode_func()
            return decode_value
        except Exception as exc:
            raise PLCDecodeError(f"Could not decode data as {data_type}") from exc

    async def connect(self):
        """
        Connect to the PLC.

        :raises PLCReadConnectionError: If the connection could not be established.
        """
        try:
            assert await self.__client.connect()
        except AssertionError as exc:
            raise PLCReadConnectionError(
                f"Could not connect to PLC (host: {self._host}, port: {self._port})"
            ) from exc

    def close(self):
        """
        Close the connection to the PLC.
        """
        self.__client.close()



"""
This module contains the constants used in the application.
"""

INFO_METRIC_NAME = "info"
CONNECTION_METRIC_NAME = "connection_up"
READ_LATENCY_METRIC_NAME = "read_time_seconds"
ERROR_FREQUENCY_METRIC_NAME = "error_count"
DEFAULT_SCRAPE_INTERVAL = "30s"
DEFAULT_PORT = 9075
DEFAULT_LOG_LEVEL = "info"
DEFAULT_NAMESPACE = "plc"
DEFAULT_IDENTIFER = "master"
DEFAULT_METRIC_LAYOUT = "dynamic"



"""
PLC Prometheus Exporter
"""

import asyncio
import time
import argparse
import math

from prometheus_client import start_http_server, Gauge, Counter, Histogram
from logger import create_logger, set_level
from config import ExporterConfig, PLCConfig, RegisterConfig, MetricLayout, load_config
from constants import (
    INFO_METRIC_NAME,
    CONNECTION_METRIC_NAME,
    ERROR_FREQUENCY_METRIC_NAME,
    READ_LATENCY_METRIC_NAME,
)
from plc_reader import (
    PLCReader,
    PLCReadError,
    PLCReadConnectionError,
    PLCRegisterType,
    PLCValueType,
    RegisterAddress,
)

prometheus_metrics: dict[str, Gauge] = {}
logger = create_logger("plc_exporter")

metric_descriptions: dict[str, str] = {
    PLCRegisterType.COILS: (
        "Coils represent discrete outputs, which are binary values "
        "and are used to control physical devices like relays, motors, lights, "
        "or any output devices connected to the PLC."
        "They can be read and written to."
    ),
    PLCRegisterType.DISCRETE_INPUTS: (
        "Discrete inputs are binary values that represent the state "
        "of physical devices like sensors, switches, or any input devices connected to the PLC. "
        "They are read-only and cannot be written to."
    ),
    PLCRegisterType.INPUT_REGISTERS: (
        "Input registers are 16-bit registers that store numeric values. "
        "They are read-only and cannot be written to."
    ),
    PLCRegisterType.HOLDING_REGISTERS: (
        "Holding registers are 16-bit registers that store numeric values. "
        "They can be read and written to."
    ),
}


def __get_metric_name(
    register_config: RegisterConfig,
    metric_layout: MetricLayout,
) -> str:
    """
    Get the name of the metric based on the layout.
    """
    if metric_layout == MetricLayout.DYNAMIC:
        return register_config.register_type.value

    return register_config.name


def __get_metric_description(
    register_config: RegisterConfig, metric_layout: MetricLayout
) -> str:
    """
    Get the description of the metric based on the layout.
    """
    if metric_layout == MetricLayout.DYNAMIC:
        return metric_descriptions[register_config.register_type]

    return register_config.description


def __get_metric_labels(
    register_config: RegisterConfig,
    metric_layout: MetricLayout,
    address: int,
    value_type: str,
    static_labels: dict,
    index: int | None = None,
) -> dict:
    """
    Get the default label set for the metric.
    """
    extra = {"index": index}

    if metric_layout == MetricLayout.STATIC:
        extra = {
            "register_type": register_config.register_type.value,
            **extra,
        }

    if metric_layout == MetricLayout.DYNAMIC:
        extra = {"name": register_config.name, **extra}

    return {
        "start_address": str(RegisterAddress(address)),
        "value_type": value_type,
        **static_labels,
        **extra,
    }


def __char_metric(
    register_config: RegisterConfig,
    value: str,
    metric_layout: MetricLayout,
    namespace: str,
    static_labels: dict,
):
    """
    Create a character metric from a single register.

    The value is converted into it's numeric ASCII representation.
    """
    name = __get_metric_name(register_config, metric_layout)
    description = __get_metric_description(register_config, metric_layout)
    labels = __get_metric_labels(
        register_config=register_config,
        metric_layout=metric_layout,
        address=register_config.address,
        value_type="ascii",
        static_labels=static_labels,
    )

    if prometheus_metrics.get(name) is None:
        prometheus_metrics[name] = Gauge(
            name=name,
            documentation=description,
            namespace=namespace,
            labelnames=list(labels.keys()),
        )

    prometheus_metrics[name].labels(**labels).set(ord(value))


def __string_metric(
    register_config: RegisterConfig,
    value: str,
    metric_layout: MetricLayout,
    namespace: str,
    static_labels: dict,
):
    """
    Create a string metric from multiple registers.

    Each character is stored in a separate metric with the same `part_of` label and
    an `index` label indicating the position in the string.

    Each character is converted into its numeric ASCII representation.
    """
    description = __get_metric_description(register_config, metric_layout)
    register_count = math.ceil(len(value) / 2)
    register_addresses = [register_config.address + i for i in range(register_count)]

    for i, char in enumerate(value):
        char_value = ord(char)
        address = register_addresses[i // 2]

        name = __get_metric_name(register_config, metric_layout)
        labels = __get_metric_labels(
            register_config=register_config,
            metric_layout=metric_layout,
            address=address,
            value_type="ascii",
            static_labels=static_labels,
            index=i,
        )

        if prometheus_metrics.get(name) is None:
            prometheus_metrics[name] = Gauge(
                name=name,
                documentation=description,
                namespace=namespace,
                labelnames=list(labels.keys()),
            )
        prometheus_metrics[name].labels(**labels).set(char_value)


def __bool_metric(
    register_config: RegisterConfig,
    value: int,
    metric_layout: MetricLayout,
    namespace: str,
    static_labels: dict,
):
    """
    Create a boolean metric from a single register.
    """
    __numeric_metric(
        register_config=register_config,
        value=int(value),
        metric_layout=metric_layout,
        namespace=namespace,
        static_labels=static_labels,
    )


def __numeric_metric(
    register_config: RegisterConfig,
    value: int,
    metric_layout: MetricLayout,
    namespace: str,
    static_labels: dict,
):
    """
    Create a metric from a numeric value stored in one or two (up to 32 bits) registers.
    """
    description = __get_metric_description(register_config, metric_layout)
    name = __get_metric_name(register_config, metric_layout)
    labels = __get_metric_labels(
        register_config=register_config,
        metric_layout=metric_layout,
        address=register_config.address,
        value_type=register_config.value_type.value,
        static_labels=static_labels,
    )

    if prometheus_metrics.get(name) is None:
        prometheus_metrics[name] = Gauge(
            name=name,
            documentation=description,
            namespace=namespace,
            labelnames=list(labels.keys()),
        )

    prometheus_metrics[name].labels(**labels).set(value)


def info_metric(namespace: str, static_labels: dict):
    """
    Create a default metric which can be relied on to filter for default
    labels created by prometheus (e.g. `job`, `instance`).
    """
    if prometheus_metrics.get(INFO_METRIC_NAME) is None:
        prometheus_metrics[INFO_METRIC_NAME] = Gauge(
            name=INFO_METRIC_NAME,
            documentation="Information about the PLC Exporter",
            namespace=namespace,
            labelnames=list(static_labels.keys()),
        )

    prometheus_metrics[INFO_METRIC_NAME].labels(**static_labels)


def connection_metric(namespace: str, static_labels: dict):
    """
    Create a default metric which can be relied on to filter for default
    labels created by prometheus (e.g. `job`, `instance`).
    """
    if prometheus_metrics.get(CONNECTION_METRIC_NAME) is None:
        prometheus_metrics[CONNECTION_METRIC_NAME] = Gauge(
            name=CONNECTION_METRIC_NAME,
            documentation="Connection status to the PLC",
            namespace=namespace,
            labelnames=list(static_labels.keys()),
        )

    prometheus_metrics[CONNECTION_METRIC_NAME].labels(**static_labels)


def set_connection_metric(
    connected: bool,
    static_labels: dict,
):
    """
    Set the connection status of the PLC.
    """
    prometheus_metrics[CONNECTION_METRIC_NAME].labels(**static_labels).set(
        int(connected)
    )


def latency_metric(namespace: str, static_labels: dict):
    """
    Create a metric to track the read latency on a register.
    """
    if prometheus_metrics.get(READ_LATENCY_METRIC_NAME) is None:
        prometheus_metrics[READ_LATENCY_METRIC_NAME] = Histogram(
            name=READ_LATENCY_METRIC_NAME,
            documentation="Read latency in milliseconds on a register",
            namespace=namespace,
            labelnames=set(
                ["name", "value_type", "start_address", "index", "register_type"]
                + list(static_labels.keys())
            ),
        )


def set_latency_metric(
    register_config: RegisterConfig,
    static_labels: dict,
    latency: float,
):
    """
    Set the latency value in milliseconds on the register that was read.
    """
    labels = __get_metric_labels(
        register_config=register_config,
        metric_layout=MetricLayout.DYNAMIC,
        address=register_config.address,
        value_type=register_config.value_type.value,
        static_labels={
            "register_type": register_config.register_type.value,
            **static_labels,
        },
    )
    prometheus_metrics[READ_LATENCY_METRIC_NAME].labels(**labels).observe(latency)


def error_metric(namespace: str, static_labels: dict):
    """
    Create a default metric to track errors.
    """
    if prometheus_metrics.get(ERROR_FREQUENCY_METRIC_NAME) is None:
        prometheus_metrics[ERROR_FREQUENCY_METRIC_NAME] = Counter(
            name=ERROR_FREQUENCY_METRIC_NAME,
            documentation="Number of errors while reading from the register",
            namespace=namespace,
            labelnames=set(
                ["name", "value_type", "start_address", "register_type", "index"]
                + list(static_labels.keys())
            ),
        )


def set_error_metric(
    register_config: RegisterConfig,
    static_labels: dict,
    init: bool = False,
):
    """
    Increment the error metric.
    """
    labels = __get_metric_labels(
        register_config=register_config,
        metric_layout=MetricLayout.DYNAMIC,
        address=register_config.address,
        value_type=register_config.value_type.value,
        static_labels={
            "register_type": register_config.register_type.value,
            **static_labels,
        },
    )
    if init is not None:
        prometheus_metrics[ERROR_FREQUENCY_METRIC_NAME].labels(**labels).inc(0)
    else:
        prometheus_metrics[ERROR_FREQUENCY_METRIC_NAME].labels(**labels).inc()


def metric(
    register_config: RegisterConfig,
    value: int,
    metric_layout: MetricLayout,
    namespace: str,
    static_labels: dict,
):
    """
    Create a metric based on the data type.
    """
    type_to_metric = {
        PLCValueType.UINT8: __numeric_metric,
        PLCValueType.UINT16: __numeric_metric,
        PLCValueType.UINT32: __numeric_metric,
        PLCValueType.UINT64: __numeric_metric,
        PLCValueType.INT8: __numeric_metric,
        PLCValueType.INT16: __numeric_metric,
        PLCValueType.INT32: __numeric_metric,
        PLCValueType.INT64: __numeric_metric,
        PLCValueType.FLOAT16: __numeric_metric,
        PLCValueType.FLOAT32: __numeric_metric,
        PLCValueType.FLOAT64: __numeric_metric,
        PLCValueType.CHAR: __char_metric,
        PLCValueType.STRING: __string_metric,
        PLCValueType.BOOL: __bool_metric,
    }

    type_to_metric[register_config.value_type](
        register_config=register_config,
        value=value,
        metric_layout=metric_layout,
        namespace=namespace,
        static_labels=static_labels,
    )


async def update_metrics(
    plc: PLCReader | None,
    mock: bool,
    metric_layout: MetricLayout,
    registers: list[RegisterConfig],
    labels: dict,
    namespace: str,
):
    """
    Update the Prometheus metrics with the values read from the PLC.
    """
    if plc is not None:
        try:
            await plc.connect()
        except PLCReadConnectionError as exc:
            set_connection_metric(connected=False, static_labels=labels)
            logger.error(exc.message)
            return

    set_connection_metric(connected=True, static_labels=labels)

    for r in registers:
        try:
            if mock:
                value = r.mock_value()
                latency = 0.125
            else:
                start_time = time.perf_counter()
                value = await plc.read(r.address, r.register_type, r.value_type, r.size)
                latency = time.perf_counter() - start_time  # seconds

            set_latency_metric(
                register_config=r,
                static_labels=labels,
                latency=latency,
            )
            metric(
                register_config=r,
                value=value,
                metric_layout=metric_layout,
                namespace=namespace,
                static_labels=labels,
            )
        except PLCReadError as exc:
            logger.error(
                "Unable to read %s with name %s: %s",
                str(RegisterAddress(r.address)),
                r.name,
                exc,
            )
            logger.debug(exc, exc_info=True)
            set_error_metric(register_config=r, static_labels=labels)
            continue

    if plc is not None:
        plc.close()


async def start_exporter(port: int, scrape_interval: int, config: dict):
    """
    Start the Prometheus exporter and update the metrics at regular intervals.
    """
    plc_config = PLCConfig.from_dict(**config["plc"])
    plc = None

    start_http_server(port=port)
    logger.info("PLC Exporter started on port %i", port)

    mock = config["mock"]
    if mock:
        logger.info("Mocking enabled. Using mock values from the config")
    else:
        plc = PLCReader(
            host=plc_config.host,
            port=plc_config.port,
            endianness=plc_config.endianness,
            word_order=plc_config.word_order,
        )

    static_labels = {
        **{"plc": config["identifier"]},
        **config["static_labels"],
    }
    all_registers = (
        config["coils"]
        + config["discrete_inputs"]
        + config["input_registers"]
        + config["holding_registers"]
    )
    register_configs = [
        RegisterConfig.from_dict(**register) for register in all_registers
    ]
    namespace = config["namespace"]

    info_metric(namespace=namespace, static_labels=static_labels)
    latency_metric(namespace=namespace, static_labels=static_labels)
    error_metric(namespace=namespace, static_labels=static_labels)
    connection_metric(namespace=namespace, static_labels=static_labels)

    for register in register_configs:
        set_error_metric(
            register_config=register, static_labels=static_labels, init=True
        )

    metric_layout = MetricLayout(config["metric_layout"])

    while True:
        logger.debug("Updating metrics")
        await update_metrics(
            plc=plc,
            mock=mock,
            metric_layout=metric_layout,
            registers=register_configs,
            labels=static_labels,
            namespace=namespace,
        )
        await asyncio.sleep(scrape_interval)


def run():
    """
    Main entry point of the exporter.
    """
    parser = argparse.ArgumentParser(description="Beckhoff PLC Prometheus Exporter")
    parser.add_argument(
        "--config",
        type=str,
        required=True,
        help="Path to YAML configuration file",
    )
    args = parser.parse_args()
    config = load_config(args.config)
    exporter_config = ExporterConfig.from_dict(**config["exporter"])

    set_level(logger, exporter_config.log_level.value)

    try:
        asyncio.run(
            start_exporter(
                port=exporter_config.port,
                scrape_interval=exporter_config.scrape_interval,
                config=config,
            )
        )
    except (KeyboardInterrupt, InterruptedError):
        logger.info("Exporter stopped")



"""
This module contains a function to create a logger.
"""

import logging
from enum import Enum


class LogLevel(Enum):
    """
    The log levels that can be used.
    """

    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL


class LowercaseLevelFormatter(logging.Formatter):
    """
    Converts the default log level to lowercase.
    """

    def format(self, record):
        record.levelname = record.levelname.lower()
        return super().format(record)


def create_logger(name: str, log_level: int = logging.DEBUG) -> logging.Logger:
    """
    Creates a console logger with the given name.
    """
    logger = logging.getLogger(name)
    logger.setLevel(log_level)
    ch = logging.StreamHandler()
    formatter = LowercaseLevelFormatter(
        'timestamp=%(asctime)s level=%(levelname)s message="%(message)s"',
        datefmt="%Y-%m-%dT%H:%M:%S",
    )
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger


def set_level(logger: logging.Logger, log_level: int) -> None:
    """
    Set the log level of the logger.
    """
    logger.setLevel(log_level)
    for handler in logger.handlers:
        handler.setLevel(log_level)



"""
This module contains the schema for the configuration file.
"""

from dataclasses import dataclass
from enum import Enum
import yaml
import humanfriendly
from schema import Schema, And, Or, Use, Optional
from logger import LogLevel
from plc_reader import PLCEndianness, PLCValueType, PLCRegisterType
from constants import (
    DEFAULT_SCRAPE_INTERVAL,
    DEFAULT_PORT,
    DEFAULT_IDENTIFER,
    DEFAULT_LOG_LEVEL,
    DEFAULT_NAMESPACE,
)


class MetricLayout(Enum):
    """
    The layout of the metrics.
    """

    DYNAMIC = "dynamic"
    STATIC = "static"


PLC_EXPORTER_SCHEMA = Schema(
    {
        Optional(
            "exporter",
            default={
                "port": DEFAULT_PORT,
                "scrape_interval": DEFAULT_SCRAPE_INTERVAL,
                "log_level": DEFAULT_LOG_LEVEL,
            },
        ): {
            Optional("port", default=DEFAULT_PORT): And(
                Use(int), lambda n: 1 <= n <= 65535
            ),
            Optional("scrape_interval", default=DEFAULT_SCRAPE_INTERVAL): str,
            Optional("log_level", default=LogLevel.INFO): And(
                Use(lambda x: x.upper()), Or(*[level.name for level in LogLevel])
            ),
        },
        "plc": {
            "host": And(str, len),
            "port": And(Use(int), lambda n: 1 <= n <= 65535),
            Optional("endianness", default=PLCEndianness.BIG.value): Or(
                *[t.value for t in PLCEndianness]
            ),
            Optional("word_order", default=PLCEndianness.BIG.value): Or(
                *[t.value for t in PLCEndianness]
            ),
        },
        Optional("namespace", default=DEFAULT_NAMESPACE): And(str, len),
        Optional("identifier", default=DEFAULT_IDENTIFER): And(str, len),
        Optional("metric_layout", default=MetricLayout.DYNAMIC): Or(
            *[t.value for t in MetricLayout]
        ),
        Optional("mock", default=False): bool,
        Optional("static_labels", default={}): dict,
        Optional("coils", default=[]): [
            {
                "name": And(str, len),
                "description": str,
                "address": And(Use(int), lambda n: n >= 0),
                Optional("value_type", default=PLCValueType.BOOL.value): Or(
                    PLCValueType.BOOL.value
                ),
                Optional("register_type", default=PLCRegisterType.COILS.value): str,
                Optional("size", default=1): And(Use(int), lambda n: n > 0),
                Optional("mock", default=0): Or(0, 1, True, False),
            }
        ],
        Optional("discrete_inputs", default=[]): [
            {
                "name": And(str, len),
                "description": str,
                "address": And(Use(int), lambda n: n >= 0),
                Optional("value_type", default=PLCValueType.BOOL.value): Or(
                    PLCValueType.BOOL.value
                ),
                Optional(
                    "register_type", default=PLCRegisterType.DISCRETE_INPUTS.value
                ): str,
                Optional("size", default=1): And(Use(int), lambda n: n > 0),
                Optional("mock", default=0): Or(0, 1, True, False),
            }
        ],
        Optional("input_registers", default=[]): [
            {
                "name": And(str, len),
                "description": str,
                "address": And(Use(int), lambda n: n >= 0),
                "value_type": Or(*[t.value for t in PLCValueType]),
                Optional(
                    "register_type", default=PLCRegisterType.INPUT_REGISTERS.value
                ): str,
                Optional("size", default=1): And(Use(int), lambda n: n > 0),
                Optional("mock", default=0): Or(int, str, bool, float),
            }
        ],
        Optional("holding_registers", default=[]): [
            {
                "name": And(str, len),
                "description": str,
                "address": And(Use(int), lambda n: n >= 0),
                "value_type": Or(*[t.value for t in PLCValueType]),
                Optional(
                    "register_type", default=PLCRegisterType.HOLDING_REGISTERS.value
                ): str,
                Optional("size", default=1): And(Use(int), lambda n: n > 0),
                Optional("mock", default=0): Or(int, str, bool, float),
            }
        ],
    }
)


class ConfigError(Exception):
    """
    Raised when there is an error in the exporter configuration.
    """


@dataclass
class RegisterConfig:
    """
    Configuration for a single register.
    """

    name: str
    description: str
    address: int
    value_type: PLCValueType
    register_type: PLCRegisterType
    size: int = 1
    mock: any = 0

    @classmethod
    def from_dict(
        cls, value_type: str, register_type: str, **kwargs
    ) -> "RegisterConfig":
        """
        Create a RegisterConfig from a dictionary.
        """
        return cls(
            value_type=PLCValueType[value_type.upper()],
            register_type=PLCRegisterType[register_type.upper()],
            **kwargs,
        )

    def register_address(self) -> str:
        """
        Return the register address as a string.
        """
        return f"0x{self.address:04x}"

    def mock_value(self) -> any:
        """
        Validates and returns the mock value.
        """
        if self.value_type in [PLCValueType.CHAR, PLCValueType.STRING]:
            assert isinstance(self.mock, str), "Mock value must be a string"

            if self.value_type == PLCValueType.CHAR:
                assert len(self.mock) == 1, "Mock value must be a single character"

            return self.mock

        if self.value_type == PLCValueType.BOOL:
            assert self.mock in [0, 1, True, False], "Mock value must be a boolean"
            return bool(self.mock)

        assert (
            self.value_type.max_value() >= self.mock >= self.value_type.min_value()
        ), (
            f"Mock value out of accepted range for type {self.value_type} "
            f"[{self.value_type.min_value()}, {self.value_type.max_value()}]"
        )
        return self.mock


@dataclass
class PLCConfig:
    """
    Configuration for the PLC.
    """

    host: str
    port: int
    endianness: PLCEndianness
    word_order: PLCEndianness

    @classmethod
    def from_dict(cls, endianness: str, word_order: str, **kwargs) -> "PLCConfig":
        """
        Create a PLCConfig from a dictionary.
        """
        return cls(
            endianness=PLCEndianness[endianness.upper()],
            word_order=PLCEndianness[word_order.upper()],
            **kwargs,
        )


@dataclass
class ExporterConfig:
    """
    Configuration for the Prometheus exporter.
    """

    port: int
    scrape_interval: int
    log_level: LogLevel

    @classmethod
    def from_dict(
        cls, scrape_interval: str, log_level: str, **kwargs
    ) -> "ExporterConfig":
        """
        Create an ExporterConfig from a dictionary.
        """
        return cls(
            scrape_interval=humanfriendly.parse_timespan(scrape_interval),
            log_level=LogLevel[log_level.upper()],
            **kwargs,
        )


def load_config(config_path: str):
    """
    Loads and validates the configuration from a YAML file.
    """
    with open(config_path, "r", encoding="utf-8") as f:
        return PLC_EXPORTER_SCHEMA.validate(yaml.safe_load(f))
